import logging
from typing import Type

import jax


def select_trainer_given_devices(
    gpus: int,
    BaseTrainer: Type,
    JittedTrainer: Type,
    MultiGPUTrainer: Type,
    disable_jit: bool,
) -> Type:
    if gpus not in [0, -1]:
        raise NotImplementedError(
            "Currently, only two options are supported: use all available (`gpus=-1`) or use no GPU (`gpus=0`)."
        )
    elif gpus == -1 and disable_jit:
        logging.warning("When training on GPUs, jit cannot be disabled.")

    if gpus == -1:
        logging.info("Training on all the available GPUs, if any.")
        trainer_cls = (
            MultiGPUTrainer
            if len([d for d in jax.devices() if d.platform == "gpu"]) > 0
            else JittedTrainer
        )

    elif gpus == 0 and disable_jit:
        logging.info("Training on CPU without jit.")
        trainer_cls = BaseTrainer
    else:
        logging.info("Training on CPU.")
        trainer_cls = JittedTrainer
    return trainer_cls



from __future__ import annotations

from typing import Optional, Tuple

import jax.numpy as jnp
from jax import random
from jax._src.prng import PRNGKeyArray

from fortuna.utils.builtins import HashableMixin


class ADVIArchitecture(HashableMixin):
    def __init__(self, dim: int, std_init_params: float = 0.1):
        """
        ADVI architecture. This consists of a simple component-wise linear transformation of the input. The
        transformation includes a mean and a log-scale parameters. With this architecture, when the base distribution
        is a diagonal Gaussian, the resulting push-forward will also be. See
         [Dinh et al., 2017](https://arxiv.org/abs/1605.08803) for reference.

        :param dim: int
            Dimension of input and output of the invertible transformation.
        :param std_init_params: float
            Standard deviation of the random Gaussian initializing the parameters.
        """
        self.dim = dim
        self.std_init_params = std_init_params

    def forward(self, params: any, u: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Component-wise forward linear transformation.

        :param params: any
            Mean and log-scale parameters.
        :param u: np.ndarray
            Input to transform forward.

        :return: Tuple[jnp.ndarray, jnp.ndarray]
            v: jnp.ndarray
                Output of the forward pass.
            ldj: jnp.ndarray
                Log-determinant of the Jacobian of the forward pass.
        """
        mean, logscale = params
        return (
            mean + jnp.exp(logscale) * u,
            jnp.repeat(jnp.sum(logscale, -1), u.shape[0]),
        )

    def backward(self, params: tuple, v: jnp.ndarray) -> tuple:
        """
        Component-wise backward linear transformation.

        :param params: tuple
            Mean and log-scale parameters.
        :param v: jnp.ndarray
            Input to transform backward.

        :return: tuple
            v: jnp.ndarray
                Output of the backward pass.
            ldj: jnp.ndarray
                Log-determinant of the Jacobian of the backward pass.
        """
        mean, logscale = params
        return (
            jnp.exp(-logscale) * (v - mean),
            jnp.repeat(-jnp.sum(logscale, -1), v.shape[0]),
        )

    def init_params(
        self, rng: PRNGKeyArray, mean: Optional[jnp.ndarray] = None,
    ) -> tuple:
        """
        Initialize mean and log-scale parameters.

        :param rng: PRNGKeyArray
            Random number generator.
        :param mean: jnp.ndarray
            If the main model has already been initialized calling `model.init`, the already
            initialized parameter values can be provided here.

        :return:
            params: tuple
                Transformation parameters.
        """
        rng, key_mean, key_logscale = random.split(rng, 3)
        if mean is None:
            mean = self.std_init_params * random.normal(key_mean, (self.dim,))
        logscale = self.std_init_params * random.normal(key_logscale, (self.dim,))
        return mean, logscale



from __future__ import annotations

import jax.numpy as jnp
from flax.core import FrozenDict

from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import Params
from fortuna.utils.strings import convert_string_to_jnp_array


class ADVIState(PosteriorState):
    """
    Attributes
    ----------
    encoded_name: jnp.ndarray
        ADVI state name encoded as an array.
    """

    encoded_name: jnp.ndarray = convert_string_to_jnp_array("ADVIState")

    @classmethod
    def convert_from_map_state(cls, map_state: MAPState, std: Params) -> ADVIState:
        """
        Convert a MAP state into an ADVI state.

        Parameters
        ----------
        map_state: MAPState
            A MAP posterior state.
        std: Params
            Standard deviation parameters.

        Returns
        -------
        ADVIState
            An ADVI state.
        """
        params = map_state.params.unfreeze()
        for k, v in std.items():
            params[k] = FrozenDict(
                {"params": dict(mean=params[k]["params"], logvar=v["params"])}
            )
        params = FrozenDict(params)
        return cls.init(
            params,
            map_state.mutable,
            map_state.tx,
            map_state.calib_params,
            map_state.calib_mutable,
        )



from typing import List, Optional, Tuple

from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.laplace import LAPLACE_NAME


class LaplacePosteriorApproximator(PosteriorApproximator):
    def __init__(self, which_params: Optional[Tuple[List, ...]] = None):
        """
        Laplace posterior approximator. It is responsible to define how the posterior distribution is approximated.

        Parameters
        ----------
        which_params: Optional[Tuple[List, ...]]
            Sequences of keys to the parameters of the probabilistic model for which to define the Laplace
            approximation. If `which_params` is not available, the Laplace approximation will be over all parameters.
        """
        if which_params:
            if type(which_params) != tuple:
                raise ValueError("`which_params` must be a tuple of lists.")
            for list_keys in which_params:
                if type(list_keys) != list:
                    raise ValueError("Each element in `which_params` must be a list.")
        self.which_params = which_params

    def __str__(self):
        return LAPLACE_NAME



from typing import Callable, Generator, Tuple

import numpy as np


def make_array_random_inputs(
    n_inputs: int, shape_inputs: Tuple[int, ...]
) -> np.ndarray:
    return np.random.normal(size=(n_inputs,) + shape_inputs)


def make_generator_fun_random_inputs(
    batch_size: int, n_batches: int, shape_inputs: Tuple[int, ...]
) -> Callable[[], Generator[np.ndarray, None, None]]:
    def inner():
        for i in range(n_batches):
            yield np.random.normal(size=(batch_size,) + shape_inputs)

    return inner


def make_generator_random_inputs(
    batch_size: int, n_batches: int, shape_inputs: Tuple[int, ...]
) -> Generator[np.ndarray, None, None]:
    for i in range(n_batches):
        yield np.random.normal(size=(batch_size,) + shape_inputs)


def make_array_random_targets(n_inputs: int, output_dim: int, output_type: str):
    if output_type not in ["discrete", "continuous"]:
        raise Exception(
            "`output_type={}` not recognized. Please choose among the following list: {}.".format(
                "discrete", "continuous"
            )
        )
    return (
        np.random.normal(size=(n_inputs, output_dim))
        if output_type == "continuous"
        else np.random.choice(output_dim, size=n_inputs)
    )


def make_generator_fun_random_targets(
    batch_size: int, n_batches: int, output_dim: int, output_type: str,
) -> Callable[[], Generator[Tuple[np.ndarray, np.ndarray], None, None]]:
    if output_type not in ["discrete", "continuous"]:
        raise Exception(
            "`output_type={}` not recognized. Please choose among the following list: {}.".format(
                "discrete", "continuous"
            )
        )

    def inner():
        for i in range(n_batches):
            yield np.random.normal(
                size=(batch_size, output_dim)
            ) if output_type == "continuous" else np.random.choice(
                output_dim, size=batch_size
            )

    return inner


def make_generator_random_targets(
    batch_size: int, n_batches: int, output_dim: int, output_type: str,
) -> Generator[Tuple[np.ndarray, np.ndarray], None, None]:
    if output_type not in ["discrete", "continuous"]:
        raise Exception(
            "`output_type={}` not recognized. Please choose among the following list: {}.".format(
                "discrete", "continuous"
            )
        )
    for i in range(n_batches):
        yield np.random.normal(
            size=(batch_size, output_dim)
        ) if output_type == "continuous" else np.random.choice(
            output_dim, size=batch_size
        )


def make_array_random_data(
    n_data: int, shape_inputs: Tuple[int, ...], output_dim: int, output_type: str
) -> Tuple[np.ndarray, np.ndarray]:
    if output_type not in ["discrete", "continuous"]:
        raise Exception(
            "`output_type={}` not recognized. Please choose among the following list: {}.".format(
                "discrete", "continuous"
            )
        )
    return (
        np.random.normal(size=(n_data,) + shape_inputs),
        np.random.normal(size=(n_data, output_dim))
        if output_type == "continuous"
        else np.random.choice(output_dim, size=n_data),
    )


def make_generator_fun_random_data(
    batch_size: int,
    n_batches: int,
    shape_inputs: Tuple[int],
    output_dim: int,
    output_type: str,
) -> Callable[[], Generator[Tuple[np.ndarray, np.ndarray], None, None]]:
    if output_type not in ["discrete", "continuous"]:
        raise Exception(
            "`output_type={}` not recognized. Please choose among the following list: {}.".format(
                "discrete", "continuous"
            )
        )

    def inner():
        for i in range(n_batches):
            yield np.random.normal(size=(batch_size,) + shape_inputs), np.random.normal(
                size=(batch_size, output_dim)
            ) if output_type == "continuous" else np.random.choice(
                output_dim, size=batch_size
            )

    return inner


def make_generator_random_data(
    batch_size: int,
    n_batches: int,
    shape_inputs: Tuple[int],
    output_dim: int,
    output_type: str,
) -> Generator[Tuple[np.ndarray, np.ndarray], None, None]:
    if output_type not in ["discrete", "continuous"]:
        raise Exception(
            "`output_type={}` not recognized. Please choose among the following list: {}.".format(
                "discrete", "continuous"
            )
        )
    for i in range(n_batches):
        yield np.random.normal(size=(batch_size,) + shape_inputs), np.random.normal(
            size=(batch_size, output_dim)
        ) if output_type == "continuous" else np.random.choice(
            output_dim, size=batch_size
        )



from typing import Callable, Optional, Tuple

import flax.linen as nn
import jax.numpy as jnp

from fortuna.typing import Array


class MLP(nn.Module):
    """
    A multi-layer perceptron (MLP).

    Parameters
    ----------
    output_dim: int
        The output model dimension.
    widths: Tuple[int]
        The number of units of each hidden layer. Default: (30, 30).
    activations: Tuple[Callable[[Array], Array]]
        The activation functions after each hidden layer. Default: (flax.linen.relu, flax.linen.relu).
    """

    output_dim: int
    widths: Optional[Tuple[int]] = (30, 30)
    activations: Optional[Tuple[Callable[[Array], Array]]] = (nn.relu, nn.relu)

    def setup(self):
        if len(self.widths) != len(self.activations):
            raise Exception(
                "`widths` and `activations` must have the same number of elements."
            )
        self.dfe_subnet = MLPDeepFeatureExtractorSubNet(
            widths=self.widths, activations=self.activations[:-1],
        )
        self.output_subnet = MLPOutputSubNet(
            activation=self.activations[-1], output_dim=self.output_dim
        )

    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        x = self.dfe_subnet(x)
        x = self.output_subnet(x)
        return x


class MLPDeepFeatureExtractorSubNet(nn.Module):
    widths: Tuple[int]
    activations: Tuple[Callable[[Array], Array]]
    """
    MLP Deep feature extractor sub-network.

    Attributes
    ----------
    widths: Tuple[int]
        The number of units of each hidden layer.
    activations: Tuple[Callable[[Array], Array]]
        The activation functions after each hidden layer. 
    """

    @nn.compact
    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        """
        Forward pass.

        Parameters
        ----------
        x: Array
            Inputs.

        Returns
        -------
        jnp.ndarray
            Output of the hidden layers.
        """
        n_activations = len(self.activations)

        def update(i: int, x):
            x = nn.Dense(self.widths[i], name="hidden" + str(i + 1))(x)
            if i < n_activations:
                x = self.activations[i](x)
            return x

        x = x.reshape(x.shape[0], -1)
        for i in range(0, len(self.widths)):
            x = update(i, x)
        return x


class MLPOutputSubNet(nn.Module):
    output_dim: int
    activation: Optional[Callable[[Array], Array]] = None
    """
    MLP output sub-network.

    Attributes
    ----------
    output_dim: int
        The output model dimension.
    activations: Tuple[Callable[[Array], Array]]
        The activation functions after each hidden layer. 
    """

    @nn.compact
    def __call__(self, x: jnp.ndarray, **kwargs) -> jnp.ndarray:
        """
        Forward pass.

        Parameters
        ----------
        x: jnp.ndarray
            Outputs of the hidden layers.

        Returns
        -------
        jnp.ndarray
            Output of the last layer.
        """
        if self.activation is not None:
            x = self.activation(x)
        x = nn.Dense(self.output_dim, name="last")(x)
        return x
